@@ -1,14 +1,17 @@
+use dns_lookup::{lookup_addr, lookup_host};
 use nix::ifaddrs::*;
 use nix::sys::socket::AddressFamily;
 use nix::sys::socket::SockaddrLike;
 use nix::sys::socket::SockaddrStorage;
 use nix::sys::stat::fstat;
 use nix::sys::stat::SFlag;
+use regex::*;
+use std::cell::RefCell;
 use std::collections::HashMap;
 use std::fmt;
+use std::io::BufRead;
 use std::net::Ipv4Addr;
 use std::net::Ipv6Addr;
-use std::net::ToSocketAddrs;
 use std::os::unix::io::RawFd;
 use std::str::FromStr;
 
@@ -28,6 +31,22 @@ pub struct NetRow {
     pub row: HashMap<String, String>,
 }
 
+#[derive(Debug, Clone)]
+pub struct Config {
+    pub interface_names: Vec<InterfaceAddress>,
+    pub hm: HashMap<String, String>,
+    pub used: Option<HashMap<Ip, bool>>,
+    pub input_family: Option<InputFamily>,
+    pub net_used: HashMap<Ip, HashMap<Ip, bool>>,
+}
+
+#[derive(Debug, Clone)]
+pub enum InputFamily {
+    Any,
+    IPv4,
+    IPv6,
+}
+
 pub enum FormatMode {
     Text,
     Binary,
@@ -109,6 +128,19 @@ impl Ip {
     }
 }
 
+pub fn host_address(ip: Addr) -> Ip {
+    match ip {
+        Addr::V4(_) => Ip {
+            address: ip,
+            cidr: 32,
+        },
+        Addr::V6(_) => Ip {
+            address: ip,
+            cidr: 128,
+        },
+    }
+}
+
 pub fn parse_mask(mask: &str) -> Option<u32> {
     match mask.parse::<u32>() {
         Ok(n) => Some(n),
@@ -118,15 +150,22 @@ pub fn parse_mask(mask: &str) -> Option<u32> {
 
 pub fn parse_v6(address: &str, input_base: Option<i32>, reverse: bool) -> Option<Addr> {
     match input_base {
-        Some(base) => Some(Addr::V6(Ipv6Addr::from(
-            match u128::from_str_radix(address, base as u32) {
-                Ok(y) => y,
-                Err(e) => {
-                    eprintln!("cannot convert {}: {}", address, e);
+        Some(base) => Some(Addr::V6(Ipv6Addr::from(if base < 0 {
+            let base = -base;
+            match i128::from_str_radix(address, base as u32) {
+                Ok(y) => y as u128,
+                Err(_e) => {
                     return None;
                 }
-            },
-        ))),
+            }
+        } else {
+            match u128::from_str_radix(address, base as u32) {
+                Ok(y) => y,
+                Err(_e) => {
+                    return None;
+                }
+            }
+        }))),
         None => match Ipv6Addr::from_str(address) {
             Ok(mut i) => {
                 if reverse {
@@ -163,15 +202,22 @@ pub fn parse_v4(address: &str, input_base: Option<i32>, reverse: bool) -> Option
                 && input_base.is_some()
                 && input_base.unwrap() != 16
             {
-                return Some(Addr::V4(Ipv4Addr::from(
-                    match u32::from_str_radix(address, base as u32) {
-                        Ok(y) => y,
-                        Err(e) => {
-                            eprintln!("cannot convert {}: {}", address, e);
+                return Some(Addr::V4(Ipv4Addr::from(if base < 0 {
+                    let base = -base;
+                    match i32::from_str_radix(address, base as u32) {
+                        Ok(y) => y as u32,
+                        Err(_e) => {
                             return None;
                         }
-                    },
-                )));
+                    }
+                } else {
+                    match u32::from_str_radix(address, base as u32) {
+                        Ok(y) => y,
+                        Err(_e) => {
+                            return None;
+                        }
+                    }
+                })));
             }
 
             let parts: Vec<String>;
@@ -192,6 +238,7 @@ pub fn parse_v4(address: &str, input_base: Option<i32>, reverse: bool) -> Option
             let mut arr: [u8; 4] = [0, 0, 0, 0];
 
             for (x, y) in split.iter().enumerate() {
+                let base = if base < 0 { -base } else { base };
                 arr[x] = match u8::from_str_radix(y, base as u32) {
                     Ok(y) => y,
                     Err(e) => {
@@ -235,6 +282,7 @@ pub fn parse_address_mask(
     default_v6_mask: Option<u32>,
     input_base: Option<i32>,
     reverse: bool,
+    config: &RefCell<Config>,
 ) -> Option<Ip> {
     let parts: Vec<&str> = a.split('/').collect();
 
@@ -247,7 +295,11 @@ pub fn parse_address_mask(
         }
     };
 
-    let input_ip = parse_v4_v6(arg, input_base, reverse);
+    let input_ip = match config.borrow().input_family {
+        Some(InputFamily::Any) | None => parse_v4_v6(arg, input_base, reverse),
+        Some(InputFamily::IPv6) => parse_v6(arg, input_base, reverse),
+        Some(InputFamily::IPv4) => parse_v4(arg, input_base, reverse),
+    };
 
     if let Some(input_ip) = input_ip {
         return Some(Ip {
@@ -267,17 +319,13 @@ pub fn parse_address_mask(
         }
     }
 
-    let addrs_iter = format!("{}:443", arg).to_socket_addrs();
-    let mut buffer: String;
-
-    if let Ok(mut address) = addrs_iter {
-        buffer = format!("{}", address.next().unwrap());
-        let v: Vec<&str> = buffer.split(':').collect();
-        buffer = v[0].to_string();
-        arg = buffer.as_str();
+    if !is_like_hostname(arg) {
+        return None;
     }
 
-    let input_ip = parse_v4_v6(arg, input_base, reverse);
+    let arg = ip_lookup(arg, &mut config.borrow_mut().hm);
+
+    let input_ip = parse_v4_v6(&arg, input_base, reverse);
 
     if let Some(input_ip) = input_ip {
         return Some(Ip {
@@ -321,16 +369,13 @@ pub fn addresses<'a>(
                         x = a
                     };
 
-                    match &used {
-                        Some(map) => {
-                            if map
-                                .get(&Addr::V4(Ipv4Addr::from(u32::from(x) - 1)))
-                                .is_some()
-                            {
-                                continue;
-                            }
+                    if let Some(map) = &used {
+                        if map
+                            .get(&Addr::V4(Ipv4Addr::from(u32::from(x) - 1)))
+                            .is_some()
+                        {
+                            continue;
                         }
-                        None => {}
                     }
 
                     return Some(if mask.is_none() {
@@ -369,16 +414,13 @@ pub fn addresses<'a>(
                         x = a
                     };
 
-                    match &used {
-                        Some(map) => {
-                            if map
-                                .get(&Addr::V6(Ipv6Addr::from(u128::from(x) - 1)))
-                                .is_some()
-                            {
-                                continue;
-                            }
+                    if let Some(map) = &used {
+                        if map
+                            .get(&Addr::V6(Ipv6Addr::from(u128::from(x) - 1)))
+                            .is_some()
+                        {
+                            continue;
                         }
-                        None => {}
                     }
 
                     return Some(if mask.is_none() {
@@ -417,21 +459,7 @@ pub fn smallest_group_network(networks: &HashMap<Ip, bool>) -> Option<Ip> {
     for key in networks.keys().skip(1) {
         let mut key = key.clone();
         match (&ip.address, &key.address) {
-            (Addr::V4(_), Addr::V4(_)) => {
-                if key.cidr < ip.cidr {
-                    ip.cidr = key.cidr;
-                }
-                key.cidr = ip.cidr;
-                while network(&key) != network(&ip) {
-                    if ip.cidr == 0 {
-                        return None;
-                    }
-                    ip.cidr -= 1;
-                    key.cidr = ip.cidr;
-                }
-                ip = network(&ip);
-            }
-            (Addr::V6(_), Addr::V6(_)) => {
+            (Addr::V4(_), Addr::V4(_)) | (Addr::V6(_), Addr::V6(_)) => {
                 if key.cidr < ip.cidr {
                     ip.cidr = key.cidr;
                 }
@@ -454,6 +482,58 @@ pub fn smallest_group_network(networks: &HashMap<Ip, bool>) -> Option<Ip> {
     Some(ip)
 }
 
+pub fn smallest_group_network_limited(networks: &HashMap<Ip, bool>, cidr: u32) -> Option<Vec<Ip>> {
+    if networks.is_empty() {
+        return None;
+    }
+
+    let mut net_list: HashMap<Ip, Ip> = HashMap::new();
+
+    for key in networks.keys() {
+        let mut bucket_ip = key.clone();
+        bucket_ip.cidr = cidr;
+
+        let mut key_copy = key.clone();
+
+        // don't add a entry with a bigger network than cidr
+        if key_copy.cidr < cidr {
+            key_copy.cidr = cidr;
+        }
+
+        let bucket = network(&bucket_ip);
+
+        if !net_list.contains_key(&bucket) {
+            net_list.insert(bucket.clone(), key_copy.clone());
+        }
+
+        let ip = net_list.get_mut(&bucket).unwrap();
+
+        match (&ip.address, &key.address) {
+            (Addr::V4(_), Addr::V4(_)) | (Addr::V6(_), Addr::V6(_)) => {
+                // println!("ip.cidr {}, key_copy.cidr {}", ip.cidr, key_copy.cidr);
+                if key_copy.cidr < ip.cidr {
+                    // println!("key less than ip: ip.cidr {}, key_copy.cidr {}", ip.cidr, key_copy.cidr);
+                    ip.cidr = key_copy.cidr;
+                }
+                key_copy.cidr = ip.cidr;
+                while network(&key_copy) != network(ip) && ip.cidr > cidr {
+                    if ip.cidr == 0 {
+                        return None;
+                    }
+                    ip.cidr -= 1;
+                    key_copy.cidr = ip.cidr;
+                }
+                *ip = network(ip);
+            }
+            (_, _) => {
+                return None;
+            }
+        }
+    }
+
+    Some(net_list.values().cloned().collect())
+}
+
 impl fmt::Display for Ip {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self.address {
@@ -1018,6 +1098,50 @@ pub fn rbl_format(ip: &Ip) -> String {
     }
 }
 
+pub fn ip_lookup(address: &str, hm: &mut HashMap<String, String>) -> String {
+    let k = format!("n/{}", address);
+
+    if let Some(v) = hm.get(&k) {
+        return v.clone();
+    }
+
+    if let Ok(buffer) = lookup_host(address) {
+        let arg = buffer[0].to_string();
+
+        hm.insert(k, arg.clone());
+        arg
+    } else {
+        "".to_string()
+    }
+}
+
+pub fn ptr_format(ip: &Ip, hm: &mut HashMap<String, String>) -> String {
+    let k: String = ip.to_string();
+
+    if let Some(v) = hm.get(&k) {
+        return v.clone();
+    }
+
+    let j = match ip.address {
+        Addr::V4(x) => {
+            let ip: std::net::IpAddr = std::net::IpAddr::V4(x);
+            lookup_addr(&ip)
+        }
+        Addr::V6(x) => {
+            let ip: std::net::IpAddr = std::net::IpAddr::V6(x);
+            lookup_addr(&ip)
+        }
+    };
+
+    if let Ok(j) = j {
+        hm.insert(k, j.clone());
+
+        return j;
+    }
+
+    "".to_string()
+}
+
 pub fn network_size(ip: &Ip) -> u128 {
     let start = network(ip);
     let end = broadcast(ip);
@@ -1046,12 +1170,16 @@ pub fn format_details(
     ip: &Ip,
     formatted: String,
     rows: &Option<HashMap<Ip, NetRow>>,
+    subnet_size: Option<u32>,
+    matches: Option<&getopts::Matches>,
+    config: &RefCell<Config>,
 ) -> Option<String> {
     let ip = &mut ip.clone();
     let mut reformatted = formatted;
 
     if rows.is_some() {
         let mut found_match = false;
+
         let rows = rows.as_ref().unwrap();
 
         match ip.address {
@@ -1120,7 +1248,11 @@ pub fn format_details(
             }
         }
         if !found_match {
-            return None;
+            if let Some(m) = matches {
+                if !m.opt_present("allowemptyrow") {
+                    return None;
+                }
+            };
         }
     }
     let b = broadcast(ip);
@@ -1132,11 +1264,6 @@ pub fn format_details(
         reformatted = reformatted.replace("%r", &r);
     }
 
-    let interfaces: Vec<InterfaceAddress> = match nix::ifaddrs::getifaddrs() {
-        Ok(x) => x.collect(),
-        Err(_) => vec![],
-    };
-
     let mut mode = FormatMode::Text;
     let mut out_str = "".to_string();
     let chars: Vec<_> = reformatted.chars().collect();
@@ -1189,17 +1316,50 @@ pub fn format_details(
                         out_str.push_str(&network_size(ip).to_string());
                     }
                     'm' => {
-                        out_str.push_str(&matching_network_interface(ip, &interfaces, false));
+                        out_str.push_str(&matching_network_interface(
+                            ip,
+                            &config.borrow().interface_names,
+                            false,
+                        ));
                     }
                     'd' => {
-                        out_str.push_str(&matching_network_interface(ip, &interfaces, true));
+                        out_str.push_str(&matching_network_interface(
+                            ip,
+                            &config.borrow().interface_names,
+                            true,
+                        ));
                     }
                     'k' => {
                         out_str.push_str(&rbl_format(ip));
                     }
+                    'p' => {
+                        out_str.push_str(&ptr_format(ip, &mut config.borrow_mut().hm));
+                    }
                     '%' => {
                         out_str.push('%');
                     }
+                    'D' => {
+                        if let Some(s) = subnet_size {
+                            out_str.push_str(&s.to_string());
+                        } else {
+                            out_str.push('D');
+                        };
+                    }
+                    'N' => {
+                        if let Some(s) = subnet_size {
+                            out_str.push_str(&subnets_in_network(s, ip).to_string());
+                        } else {
+                            out_str.push('N');
+                        };
+                    }
+                    'C' => {
+                        if let Some(used) = &config.borrow().used {
+                            let count = used.keys().count();
+                            out_str.push_str(&count.to_string());
+                        } else {
+                            out_str.push('C');
+                        }
+                    }
                     _ => {
                         out_str.push(k);
                     }
@@ -1265,3 +1425,91 @@ pub fn fd_ready(fd: RawFd) -> bool {
 
     false
 }
+
+pub fn find_ips<'a>(
+    reader: &'a mut Box<dyn BufRead>,
+    input_base: Option<i32>,
+    reverse: &'a Reverse,
+    config: &'a RefCell<Config>,
+) -> impl 'a + std::iter::Iterator<Item = Vec<Ip>> {
+    std::iter::from_fn(move || {
+        if let Some(line) = reader.lines().next().into_iter().by_ref().next() {
+            let line: String = line.as_ref().unwrap().trim().to_string();
+            let mut v = vec![];
+
+            for part in line.split(' ') {
+                let p = part.trim();
+                if p.is_empty() {
+                    continue;
+                }
+
+                let ip = match parse_address_mask(
+                    p,
+                    Some(32),
+                    Some(128),
+                    input_base,
+                    matches!(reverse, Reverse::Both | Reverse::Input),
+                    config,
+                ) {
+                    Some(x) => x,
+                    None => {
+                        eprintln!("Could not parse {}", p);
+                        continue;
+                    }
+                };
+
+                v.push(ip);
+            }
+            return Some(v);
+        }
+        None
+    })
+}
+
+pub fn subnets_in_network(networks: u32, ip: &Ip) -> u128 {
+    let base: u128 = 2;
+    base.pow(networks - ip.cidr)
+}
+
+pub fn is_like_hostname(word: &str) -> bool {
+    let r = Regex::new(r#"^(\S+\.)?(xn--)?[a-z0-9][a-z0-9-_]{0,61}[a-z0-9]{0,1}\.(xn--)?([a-z0-9\-]{1,61}|[a-z0-9-]{1,30}\.[a-z]{2,})$"#).unwrap();
+
+    r.is_match(word)
+}
+
+pub fn inside_filter(inside: Option<bool>, ip_args: &[Ip], ip: &Ip) -> bool {
+    match inside {
+        Some(true) => {
+            let mut found = false;
+            for arg in ip_args {
+                if within(arg, ip) {
+                    found = true;
+                    break;
+                }
+            }
+
+            if found {
+                return true;
+            }
+        }
+        Some(false) => {
+            let mut found = false;
+
+            for arg in ip_args {
+                if within(arg, ip) {
+                    found = true;
+                    break;
+                }
+            }
+
+            if !found {
+                return true;
+            }
+        }
+        None => {
+            return true;
+        }
+    }
+
+    false
+}
