@@ -1,5 +1,6 @@
 use getopts::Options;
 use ripcalc::*;
+use std::cell::RefCell;
 use std::collections::HashMap;
 use std::fs::File;
 use std::io::{BufRead, BufReader};
@@ -11,14 +12,48 @@ fn print_details(
     matches: &getopts::Matches,
     rows: &Option<HashMap<Ip, NetRow>>,
     used: Option<&HashMap<Addr, bool>>,
+    config: &RefCell<Config>,
 ) {
+    let mut networks: Option<u32> = None;
+
+    if matches.opt_present("networks") {
+        let nets = matches.opt_str("networks").unwrap().trim().parse().unwrap();
+
+        if nets < ip.cidr && !(matches.opt_present("encapsulating") && matches.opt_present("group"))
+        {
+            eprintln!("{} is bigger than the network mask {}", nets, ip.cidr);
+            std::process::exit(1);
+        }
+
+        match ip.address {
+            Addr::V4(_) => {
+                if nets > 32 {
+                    eprintln!("{} is too big", nets);
+                    std::process::exit(1);
+                }
+            }
+            Addr::V6(_) => {
+                if nets > 128 {
+                    eprintln!("{} is too big", nets);
+                    std::process::exit(1);
+                }
+            }
+        }
+        networks = Some(nets);
+    }
+
     let mut formatted = if matches.opt_present("f") {
         matches.opt_str("f").unwrap()
     } else {
+        let mut network_size = "Network size: %t".to_string();
         let width = 25;
+        if matches.opt_present("networks") {
+            network_size = format!("Networks ({}): %N", networks.as_ref().unwrap()).to_string();
+        }
         match ip.address {
-            Addr::V4(_) => format!("{ip:>width$}/{cidr}\n{broadcast:>width$}\n{network:>width$}\n{subnet:>width$}\n{wildcard:>width$}\n{network_size:>width$}\n", ip="IP is: %a", cidr="%c", broadcast="Broadcast is: %b", network="Network is: %n", subnet="Subnet is: %s", wildcard="Wildcard is: %w", network_size="Network size: %t", width=width),
-            Addr::V6(_) => format!("{ip:>widthn$}/{cidr}\n{expanded:>width$}\n{network:>width$}\n{last_host_address:>width$}\n{subnet:>width$}\n{network_size:>widthn$}\n", ip="IP is: %a", cidr="%c", expanded="Expanded: %xa", network="Network is: %xn", last_host_address="Last host address: %xb", subnet="Subnet is: %xs", network_size="Network size: %t", width=width, widthn=width-1),
+
+            Addr::V4(_) => format!("{ip:>width$}/{cidr}\n{broadcast:>width$}\n{network:>width$}\n{subnet:>width$}\n{wildcard:>width$}\n{network_size:>width$}\n", ip="IP is: %a", cidr="%c", broadcast="Broadcast is: %b", network="Network is: %n", subnet="Subnet is: %s", wildcard="Wildcard is: %w", network_size=network_size, width=width),
+            Addr::V6(_) => format!("{ip:>widthn$}/{cidr}\n{expanded:>width$}\n{network:>width$}\n{last_host_address:>width$}\n{subnet:>width$}\n{network_size:>widthn$}\n", ip="IP is: %a", cidr="%c", expanded="Expanded: %xa", network="Network is: %xn", last_host_address="Last host address: %xb", subnet="Subnet is: %xs", network_size=network_size, width=width, widthn=width-1),
         }
     };
 
@@ -40,7 +75,14 @@ fn print_details(
         };
 
         for ip_copy in addresses(ip, used, Some(divide)) {
-            if let Some(m) = format_details(&ip_copy, formatted.to_string(), rows) {
+            if let Some(m) = format_details(
+                &ip_copy,
+                formatted.to_string(),
+                rows,
+                networks,
+                Some(matches),
+                config,
+            ) {
                 print!("{}", m);
             }
         }
@@ -49,21 +91,28 @@ fn print_details(
 
     if matches.opt_present("list") {
         if matches.opt_present("noexpand") {
-            if let Some(m) = format_details(ip, formatted, rows) {
+            if let Some(m) = format_details(ip, formatted, rows, networks, Some(matches), config) {
                 print!("{}", m);
             }
             return;
         }
 
         for ip_copy in addresses(ip, used, None) {
-            if let Some(m) = format_details(&ip_copy, formatted.to_string(), rows) {
+            if let Some(m) = format_details(
+                &ip_copy,
+                formatted.to_string(),
+                rows,
+                networks,
+                Some(matches),
+                config,
+            ) {
                 print!("{}", m);
             }
         }
         return;
     }
 
-    if let Some(m) = format_details(ip, formatted, rows) {
+    if let Some(m) = format_details(ip, formatted, rows, networks, Some(matches), config) {
         print!("{}", m);
     }
 }
@@ -180,6 +229,86 @@ fn process_csv(
     }
 }
 
+fn process_encapsulated_input(
+    matches: &getopts::Matches,
+    inside: Option<bool>,
+    ip_args: &[Ip],
+    i: &Ip,
+    network_size: u32,
+    used: &mut HashMap<Ip, bool>,
+    config: &RefCell<Config>,
+) {
+    if !inside_filter(inside, ip_args, i) {
+        return;
+    }
+
+    used.insert(i.clone(), true);
+    if matches.opt_present("group") {
+        let net_mask = network(&Ip {
+            address: i.address.clone(),
+            cidr: network_size,
+        });
+        let nu = &mut config.borrow_mut().net_used;
+
+        if nu.get(&net_mask).is_none() {
+            nu.insert(net_mask.clone(), HashMap::new());
+        }
+
+        let nm = nu.get_mut(&net_mask).unwrap();
+        nm.insert(i.clone(), true);
+    }
+}
+
+fn process_group_encapsulation(
+    matches: &getopts::Matches,
+    used: HashMap<Ip, bool>,
+    rows: &Option<HashMap<Ip, NetRow>>,
+    config: &RefCell<Config>,
+) {
+    if matches.opt_present("group") {
+        let network_size: u32 = matches.opt_str("group").unwrap().trim().parse().unwrap();
+        match smallest_group_network_limited(&used, network_size) {
+            Some(mut x) => {
+                x.sort_by(|a, b| a.partial_cmp(b).unwrap());
+                for y in x {
+                    {
+                        let net_mask = network(&Ip {
+                            address: y.address.clone(),
+                            cidr: network_size,
+                        });
+
+                        let cfg = &mut config.borrow_mut();
+                        if !cfg.net_used.contains_key(&net_mask) {
+                            cfg.net_used.insert(net_mask.clone(), HashMap::new());
+                        }
+
+                        let nm = cfg.net_used.get(&net_mask).unwrap();
+
+                        cfg.used = Some(nm.clone());
+                    }
+
+                    print_details(&y, matches, rows, None, config);
+                }
+            }
+            None => {
+                eprintln!("Could not find an encapsulating network, sorry");
+                std::process::exit(1);
+            }
+        }
+    } else {
+        match smallest_group_network(&used) {
+            Some(x) => {
+                print_details(&x, matches, rows, None, config);
+            }
+            None => {
+                eprintln!("Could not find an encapsulating network, sorry");
+                std::process::exit(1);
+            }
+        }
+    }
+}
+
+#[allow(clippy::too_many_arguments)]
 fn process_input_file(
     path: &str,
     matches: &getopts::Matches,
@@ -188,8 +317,9 @@ fn process_input_file(
     ip_args: &[Ip],
     rows: &Option<HashMap<Ip, NetRow>>,
     inside: Option<bool>,
+    config: &RefCell<Config>,
 ) {
-    let reader: Box<dyn BufRead> = if path == "-" {
+    let mut reader: Box<dyn BufRead> = if path == "-" {
         Box::new(BufReader::new(std::io::stdin()))
     } else {
         let path = std::path::Path::new(&path);
@@ -203,114 +333,56 @@ fn process_input_file(
         Box::new(BufReader::new(File::open(path).unwrap()))
     };
 
-    let mut used: HashMap<Addr, bool> = HashMap::new();
     if matches.opt_present("available") {
-        for line in reader.lines() {
-            let line: String = line.as_ref().unwrap().trim().to_string();
-            let ip = match parse_address_mask(
-                line.as_ref(),
-                Some(32),
-                Some(128),
-                input_base,
-                matches!(reverse, Reverse::Both | Reverse::Input),
-            ) {
-                Some(x) => x,
-                None => {
-                    eprintln!("Could not parse {}", line);
-                    continue;
-                }
-            };
-            for ip in addresses(&ip, None, None) {
+        let mut used: HashMap<Addr, bool> = HashMap::new();
+        for a in find_ips(&mut reader, input_base, reverse, config) {
+            for ip in a {
                 used.insert(ip.address, true);
             }
         }
 
         for arg in ip_args {
-            print_details(arg, matches, rows, Some(&used));
+            print_details(arg, matches, rows, Some(&used), config);
         }
         std::process::exit(0);
     }
 
     if matches.opt_present("encapsulating") {
         let mut used: HashMap<Ip, bool> = HashMap::new();
-        for line in reader.lines() {
-            let line: String = line.as_ref().unwrap().trim().to_string();
-            let ip = match parse_address_mask(
-                line.as_ref(),
-                Some(32),
-                Some(128),
-                input_base,
-                matches!(reverse, Reverse::Both | Reverse::Input),
-            ) {
-                Some(x) => x,
-                None => {
-                    eprintln!("Could not parse {}", line);
-                    continue;
-                }
-            };
-            used.insert(ip, true);
+        let mut network_size: u32 = 0;
+
+        if matches.opt_present("group") {
+            network_size = matches.opt_str("group").unwrap().trim().parse().unwrap();
         }
 
-        match smallest_group_network(&used) {
-            Some(x) => {
-                print_details(&x, matches, rows, None);
-            }
-            None => {
-                eprintln!("Could not find an encapsulating network, sorry");
-                std::process::exit(1);
+        for a in find_ips(&mut reader, input_base, reverse, config) {
+            for i in a {
+                process_encapsulated_input(
+                    matches,
+                    inside,
+                    ip_args,
+                    &i,
+                    network_size,
+                    &mut used,
+                    config,
+                );
             }
         }
 
+        config.borrow_mut().used = Some(used.clone());
+
+        process_group_encapsulation(matches, used, rows, config);
+
         std::process::exit(0);
     }
 
     let mut found_match = false;
 
-    for line in reader.lines() {
-        let line: String = line.as_ref().unwrap().trim().to_string();
-        let ip = parse_address_mask(
-            &line,
-            Some(32),
-            Some(128),
-            input_base,
-            matches!(reverse, Reverse::Both | Reverse::Source),
-        );
-        if ip.is_none() {
-            continue;
-        }
-
-        match inside {
-            Some(true) => {
-                let mut found = false;
-                for arg in ip_args {
-                    if within(arg, ip.as_ref().unwrap()) {
-                        found = true;
-                        break;
-                    }
-                }
-
-                if found {
-                    found_match = true;
-                    print_details(&ip.unwrap(), matches, rows, None);
-                }
-            }
-            Some(false) => {
-                let mut found = false;
-
-                for arg in ip_args {
-                    if within(arg, ip.as_ref().unwrap()) {
-                        found = true;
-                        break;
-                    }
-                }
-
-                if !found {
-                    found_match = true;
-                    print_details(&ip.unwrap(), matches, rows, None);
-                }
-            }
-            None => {
-                print_details(&ip.unwrap(), matches, rows, None);
+    for a in find_ips(&mut reader, input_base, reverse, config) {
+        for ip in a {
+            if inside_filter(inside, ip_args, &ip) {
+                found_match = true;
+                print_details(&ip, matches, rows, None, config);
             }
         }
     }
@@ -339,18 +411,31 @@ fn wait_stdin(matches: &getopts::Matches) -> bool {
 fn main() {
     let mut opts = Options::new();
     let mut rows: Option<HashMap<Ip, NetRow>> = None;
-    let mut input_ip: Option<Addr> = None;
+    let input_ip: Option<Addr> = None;
     let mut input_mask: Option<u32> = None;
     let mut input_base: Option<i32> = None;
     let mut reverse = Reverse::None;
     let mut inside: Option<bool> = None;
     let args: Vec<String> = std::env::args().collect();
     let mut ip_args: Vec<Ip> = vec![];
+    let config = RefCell::new(Config {
+        interface_names: vec![],
+        hm: HashMap::new(),
+        used: None,
+        input_family: None,
+        net_used: HashMap::new(),
+    });
+
     opts.parsing_style(getopts::ParsingStyle::FloatingFrees);
-    opts.optopt("4", "ipv4", "ipv4 address", "IPv4");
-    opts.optopt("6", "ipv6", "ipv6 address", "IPv6");
+    opts.optflag("4", "ipv4", "treat inputs as ipv4 address");
+    opts.optflag("6", "ipv6", "treat inputs as ipv6 address");
 
     opts.optflag("a", "available", "display unused addresses");
+    opts.optflag(
+        "",
+        "allowemptyrow",
+        "when no matching csv network, use empty fields",
+    );
     opts.optopt("b", "base", "ipv4 base format, default to oct", "INTEGER");
     opts.optopt("c", "csv", "csv reference file", "PATH");
     opts.optopt("d", "divide", "divide network into chunks", "CIDR");
@@ -363,6 +448,12 @@ fn main() {
     );
 
     opts.optopt("f", "format", "format output\n'cidr' expands to %a/%c\\n\n'short' expands to %a\\n\nSee manual for more options", "STRING");
+    opts.optopt(
+        "",
+        "group",
+        "maximum network group size for encapsulation",
+        "CIDR",
+    );
     opts.optflag("h", "help", "display help");
 
     opts.optopt("i", "field", "csv field", "FIELD");
@@ -374,6 +465,12 @@ fn main() {
     );
     opts.optflag("", "inside", "display when extremities are inside network");
     opts.optopt("m", "mask", "cidr mask", "CIDR");
+    opts.optopt(
+        "n",
+        "networks",
+        "instead of hosts, display number of subnets of this size",
+        "CIDR",
+    );
 
     opts.optopt(
         "r",
@@ -416,6 +513,26 @@ fn main() {
         inside = Some(false);
     }
 
+    if matches.opt_present("group") {
+        let _: u32 = match matches.opt_str("group").unwrap().trim().parse() {
+            Ok(x) => x,
+            Err(x) => {
+                eprintln!("Cannot convert {} to number", x);
+                std::process::exit(1);
+            }
+        };
+    }
+
+    if matches.opt_present("networks") {
+        let _: u32 = match matches.opt_str("networks").unwrap().trim().parse() {
+            Ok(x) => x,
+            Err(x) => {
+                eprintln!("Cannot convert {} to number", x);
+                std::process::exit(1);
+            }
+        };
+    }
+
     if matches.opt_present("reverse") {
         match matches.opt_str("reverse").unwrap().as_str() {
             "inputs" => {
@@ -435,7 +552,7 @@ fn main() {
     }
 
     if matches.opt_present("base") {
-        input_base = match i32::from_str(&matches.opt_str("b").unwrap()) {
+        input_base = match i32::from_str(&matches.opt_str("base").unwrap()) {
             Ok(x) => Some(x),
             Err(x) => {
                 println!("Cannot convert to an integer base: {}", x);
@@ -466,20 +583,12 @@ fn main() {
         input_mask = parse_mask(&v);
     }
 
-    if let Some(v) = matches.opt_str("ipv4") {
-        input_ip = parse_v4(
-            &v,
-            input_base,
-            matches!(reverse, Reverse::Both | Reverse::Input),
-        );
+    if matches.opt_present("ipv4") {
+        config.borrow_mut().input_family = Some(InputFamily::IPv4);
     }
 
-    if let Some(v) = matches.opt_str("ipv6") {
-        input_ip = parse_v6(
-            &v,
-            input_base,
-            matches!(reverse, Reverse::Both | Reverse::Input),
-        );
+    if matches.opt_present("ipv6") {
+        config.borrow_mut().input_family = Some(InputFamily::IPv6);
     }
 
     if input_mask.is_none() && matches.free.is_empty() {
@@ -506,6 +615,7 @@ fn main() {
                 input_mask,
                 input_base,
                 matches!(reverse, Reverse::Both | Reverse::Input),
+                &config,
             );
 
             if let Some(ip) = ip {
@@ -522,7 +632,7 @@ fn main() {
             matches.opt_str("file").unwrap()
         };
         process_input_file(
-            &path, &matches, input_base, &reverse, &ip_args, &rows, inside,
+            &path, &matches, input_base, &reverse, &ip_args, &rows, inside, &config,
         );
 
         if ip_args.clone().is_empty() || inside.is_some() {
@@ -549,23 +659,30 @@ fn main() {
         }
 
         if matches.opt_present("encapsulating") {
+            let mut network_size: u32 = 0;
+
+            if matches.opt_present("group") {
+                network_size = matches.opt_str("group").unwrap().trim().parse().unwrap();
+            }
+            process_encapsulated_input(
+                &matches,
+                inside,
+                &ip_args,
+                arg,
+                network_size,
+                &mut used,
+                &config,
+            );
             used.insert(arg.clone(), true);
             continue;
         }
-        print_details(arg, &matches, &rows, None);
+        print_details(arg, &matches, &rows, None, &config);
     }
 
-    if matches.opt_present("encapsulating") {
-        match smallest_group_network(&used) {
-            Some(x) => {
-                print_details(&x, &matches, &rows, None);
-            }
-            None => {
-                eprintln!("Could not find an encapsulating network, sorry");
-                std::process::exit(1);
-            }
-        }
+    config.borrow_mut().used = Some(used.clone());
 
+    if matches.opt_present("encapsulating") {
+        process_group_encapsulation(&matches, used, &rows, &config);
         std::process::exit(0);
     }
 
