@@ -1,3 +1,4 @@
+use std::cell::RefCell;
 use std::collections::HashMap;
 use std::net::Ipv4Addr;
 use std::net::Ipv6Addr;
@@ -270,20 +271,32 @@ mod test {
         assert_eq!(i.next(), None);
     }
 
+    fn default_config() -> Config {
+        Config {
+            interface_names: vec![],
+            hm: HashMap::new(),
+            used: None,
+            input_family: None,
+            net_used: HashMap::new(),
+        }
+    }
+
     #[test]
     fn test_format_ng_ip() {
         let net = Ip {
             address: Addr::V4(Ipv4Addr::from_str("192.168.0.0").unwrap()),
             cidr: 30,
         };
+        let config = RefCell::new(default_config());
 
-        let f = format_details(&net, "%a".to_string(), &None);
+        let f = format_details(&net, "%a".to_string(), &None, None, None, &config);
 
         assert_eq!(f, Some("192.168.0.0".to_string()));
     }
 
     #[test]
     fn test_format_ng_percent() {
+        let config = RefCell::new(default_config());
         let f = format_details(
             &Ip {
                 address: Addr::V4(Ipv4Addr::from_str("192.168.0.0").unwrap()),
@@ -291,6 +304,9 @@ mod test {
             },
             "%".to_string(),
             &None,
+            None,
+            None,
+            &config,
         );
 
         assert_eq!(f, Some("%".to_string()));
@@ -302,6 +318,9 @@ mod test {
             },
             "%%".to_string(),
             &None,
+            None,
+            None,
+            &config,
         );
 
         assert_eq!(f, Some("%".to_string()));
@@ -313,6 +332,9 @@ mod test {
             },
             "%%%".to_string(),
             &None,
+            None,
+            None,
+            &config,
         );
 
         assert_eq!(f, Some("%%".to_string()));
@@ -324,6 +346,9 @@ mod test {
             },
             "%%%%".to_string(),
             &None,
+            None,
+            None,
+            &config,
         );
 
         assert_eq!(f, Some("%%".to_string()));
@@ -336,7 +361,8 @@ mod test {
             cidr: 64,
         };
 
-        let f = format_details(&net, "select * from IP6 where (ip >= %ln and ip <= %lb) and active = 1;\nupdate IP6 set active = 0 where (ip >= %ln and ip <= %lb) and active = 1;".to_string(), &None);
+        let config = RefCell::new(default_config());
+        let f = format_details(&net, "select * from IP6 where (ip >= %ln and ip <= %lb) and active = 1;\nupdate IP6 set active = 0 where (ip >= %ln and ip <= %lb) and active = 1;".to_string(), &None, None, None, &config);
 
         assert_eq!(f, Some("select * from IP6 where (ip >= 42540724579414763292693624807812497408 and ip <= 42540724579414763311140368881522049023) and active = 1;
 update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 and ip <= 42540724579414763311140368881522049023) and active = 1;".to_string()));
@@ -349,7 +375,8 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
             cidr: 64,
         };
 
-        let f = format_details(&net, "%%b".to_string(), &None);
+        let config = RefCell::new(default_config());
+        let f = format_details(&net, "%%b".to_string(), &None, None, None, &config);
 
         assert_eq!(f, Some("%b".to_string()));
     }
@@ -361,7 +388,8 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
             cidr: 64,
         };
 
-        let f = format_details(&net, "%lb".to_string(), &None);
+        let config = RefCell::new(default_config());
+        let f = format_details(&net, "%lb".to_string(), &None, None, None, &config);
 
         assert_eq!(
             f,
@@ -376,7 +404,8 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
             cidr: 64,
         };
 
-        let f = format_details(&net, "%lb\n\n\n%%".to_string(), &None);
+        let config = RefCell::new(default_config());
+        let f = format_details(&net, "%lb\n\n\n%%".to_string(), &None, None, None, &config);
 
         assert_eq!(
             f,
@@ -390,17 +419,18 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
             address: Addr::V6(Ipv6Addr::from_str("2001:ba8:1f1:f1cb::4").unwrap()),
             cidr: 64,
         };
+        let config = RefCell::new(default_config());
 
-        let f = format_details(&net, "\n".to_string(), &None);
+        let f = format_details(&net, "\n".to_string(), &None, None, None, &config);
         assert_eq!(f, Some("\n".to_string()));
 
-        let f = format_details(&net, "\\".to_string(), &None);
+        let f = format_details(&net, "\\".to_string(), &None, None, None, &config);
         assert_eq!(f, Some('\\'.to_string()));
 
-        let f = format_details(&net, "\\i".to_string(), &None);
+        let f = format_details(&net, "\\i".to_string(), &None, None, None, &config);
         assert_eq!(f, Some("i".to_string()));
 
-        let f = format_details(&net, "\\t".to_string(), &None);
+        let f = format_details(&net, "\\t".to_string(), &None, None, None, &config);
         assert_eq!(f, Some("\t".to_string()));
     }
 
@@ -577,8 +607,9 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
 
     #[test]
     fn test_base_hex() {
+        let config = RefCell::new(default_config());
         assert_eq!(
-            parse_address_mask("192.168.1.1", None, None, Some(10), false),
+            parse_address_mask("192.168.1.1", None, None, Some(10), false, &config),
             Some(Ip {
                 address: Addr::V4(Ipv4Addr::from_str("192.168.1.1").unwrap()),
                 cidr: 24,
@@ -586,21 +617,21 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
         );
 
         assert_eq!(
-            parse_address_mask("192.168.1.1", None, None, None, false),
+            parse_address_mask("192.168.1.1", None, None, None, false, &config),
             Some(Ip {
                 address: Addr::V4(Ipv4Addr::from_str("192.168.1.1").unwrap()),
                 cidr: 24,
             })
         );
         assert_eq!(
-            parse_address_mask("D4166001", None, None, Some(16), false),
+            parse_address_mask("D4166001", None, None, Some(16), false, &config),
             Some(Ip {
                 address: Addr::V4(Ipv4Addr::from_str("212.22.96.1").unwrap()),
                 cidr: 24,
             })
         );
         assert_eq!(
-            parse_address_mask("177.0.0.1", None, None, Some(8), false),
+            parse_address_mask("177.0.0.1", None, None, Some(8), false, &config),
             Some(Ip {
                 address: Addr::V4(Ipv4Addr::from_str("127.0.0.1").unwrap()),
                 cidr: 24,
@@ -610,8 +641,10 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
 
     #[test]
     fn test_reverse() {
+        let config = RefCell::new(default_config());
+
         assert_eq!(
-            parse_address_mask("0101A8C0", None, None, Some(16), true),
+            parse_address_mask("0101A8C0", None, None, Some(16), true, &config),
             Some(Ip {
                 address: Addr::V4(Ipv4Addr::from_str("192.168.1.1").unwrap()),
                 cidr: 24,
@@ -626,9 +659,11 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
             cidr: 30,
         };
 
-        let f = format_details(&net, "%La".to_string(), &None);
+        let mut config = RefCell::new(default_config());
+
+        let f = format_details(&net, "%La".to_string(), &None, None, None, &mut config);
         assert_eq!(f, Some("-1819047474".to_string()));
-        let f = format_details(&net, "%la".to_string(), &None);
+        let f = format_details(&net, "%la".to_string(), &None, None, None, &mut config);
         assert_eq!(f, Some("2475919822".to_string()));
 
         let net = Ip {
@@ -636,9 +671,9 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
             cidr: 30,
         };
 
-        let f = format_details(&net, "%La".to_string(), &None);
+        let f = format_details(&net, "%La".to_string(), &None, None, None, &mut config);
         assert_eq!(f, Some("-5192296858534827628530496329154561".to_string()));
-        let f = format_details(&net, "%la".to_string(), &None);
+        let f = format_details(&net, "%la".to_string(), &None, None, None, &mut config);
         assert_eq!(
             f,
             Some("340277174624079928635746076935439056895".to_string())
@@ -806,4 +841,189 @@ update IP6 set active = 0 where (ip >= 42540724579414763292693624807812497408 an
         let v: Vec<Ip> = i.collect();
         assert_eq!(v.len(), 65536);
     }
+
+    #[test]
+    fn test_networks_sizing_v6() {
+        let net = Ip {
+            address: Addr::V6(Ipv6Addr::from_str("::1").unwrap()),
+            cidr: 48,
+        };
+
+        assert_eq!(subnets_in_network(64, &net), 65536);
+        assert_eq!(subnets_in_network(48, &net), 1);
+        assert_eq!(subnets_in_network(49, &net), 2);
+    }
+
+    #[test]
+    fn test_networks_sizing_v4() {
+        let net = Ip {
+            address: Addr::V4(Ipv4Addr::from_str("127.0.0.1").unwrap()),
+            cidr: 16,
+        };
+
+        assert_eq!(subnets_in_network(24, &net), 256);
+        assert_eq!(subnets_in_network(25, &net), 512);
+        assert_eq!(subnets_in_network(26, &net), 1024);
+        assert_eq!(subnets_in_network(27, &net), 2048);
+        assert_eq!(subnets_in_network(28, &net), 4096);
+    }
+
+    #[test]
+    fn test_smallest_network_limited() {
+        let empty: HashMap<Ip, bool> = HashMap::new();
+        assert_eq!(smallest_group_network_limited(&empty, 32), None);
+    }
+
+    #[test]
+    fn test_smallest_network_limited_22_24() {
+        let mut net_list: HashMap<Ip, bool> = HashMap::new();
+        for i in 0..4 {
+            for j in 0..4 {
+                net_list.insert(
+                    Ip {
+                        address: Addr::V4(
+                            Ipv4Addr::from_str(&format!("192.168.{i}.{j}")).expect("bad ipv4"),
+                        ),
+                        cidr: 24,
+                    },
+                    true,
+                );
+            }
+        }
+
+        let mut resp = smallest_group_network_limited(&net_list, 22).unwrap();
+        resp.sort_by(|a, b| a.partial_cmp(b).unwrap());
+        assert_eq!(
+            resp,
+            [Ip {
+                address: Addr::V4(Ipv4Addr::from_str(&format!("192.168.0.0")).expect("bad ipv4")),
+                cidr: 22
+            }]
+        );
+    }
+
+    #[test]
+    fn test_smallest_network_limited_22_8() {
+        let mut net_list: HashMap<Ip, bool> = HashMap::new();
+        for i in 0..4 {
+            for j in 0..4 {
+                net_list.insert(
+                    Ip {
+                        address: Addr::V4(
+                            Ipv4Addr::from_str(&format!("192.{i}.{j}.0")).expect("bad ipv4"),
+                        ),
+                        cidr: 8,
+                    },
+                    true,
+                );
+            }
+        }
+
+        let mut resp = smallest_group_network_limited(&net_list, 22).unwrap();
+        resp.sort_by(|a, b| a.partial_cmp(b).unwrap());
+        assert_eq!(
+            resp,
+            [
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.0.0.0")).expect("bad ipv4")),
+                    cidr: 22
+                },
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.1.0.0")).expect("bad ipv4")),
+                    cidr: 22
+                },
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.2.0.0")).expect("bad ipv4")),
+                    cidr: 22
+                },
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.3.0.0")).expect("bad ipv4")),
+                    cidr: 22
+                },
+            ]
+        );
+    }
+
+    #[test]
+    fn test_smallest_network_limited_24_8() {
+        let mut net_list: HashMap<Ip, bool> = HashMap::new();
+        for i in 0..4 {
+            for j in 0..4 {
+                net_list.insert(
+                    Ip {
+                        address: Addr::V4(
+                            Ipv4Addr::from_str(&format!("192.0.{i}.{j}")).expect("bad ipv4"),
+                        ),
+                        cidr: 8,
+                    },
+                    true,
+                );
+            }
+        }
+
+        let mut resp = smallest_group_network_limited(&net_list, 24).unwrap();
+        resp.sort_by(|a, b| a.partial_cmp(b).unwrap());
+        assert_eq!(
+            resp,
+            [
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.0.0.0")).expect("bad ipv4")),
+                    cidr: 24
+                },
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.0.1.0")).expect("bad ipv4")),
+                    cidr: 24
+                },
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.0.2.0")).expect("bad ipv4")),
+                    cidr: 24
+                },
+                Ip {
+                    address: Addr::V4(Ipv4Addr::from_str(&format!("192.0.3.0")).expect("bad ipv4")),
+                    cidr: 24
+                },
+            ]
+        );
+    }
+
+    #[test]
+    fn test_is_like_hostname() {
+        assert_eq!(is_like_hostname("www.usenix.org.uk"), true);
+        assert_eq!(is_like_hostname("www.usenix.org.uk:80"), false);
+        assert_eq!(is_like_hostname("org.uk"), true);
+        assert_eq!(is_like_hostname("bbc.co.uk"), true);
+        assert_eq!(is_like_hostname("org"), false);
+        assert_eq!(is_like_hostname("n"), false);
+        assert_eq!(is_like_hostname("n!"), false);
+        assert_eq!(is_like_hostname("http://www.bbc.co.uk/"), false);
+    }
+
+    #[test]
+    fn test_inside_filter() {
+        let config = RefCell::new(default_config());
+
+        let mut net_list = vec![];
+
+        net_list
+            .push(parse_address_mask("192.168.0.0/24", None, None, None, false, &config).unwrap());
+
+        let ip = parse_address_mask("192.168.0.0/24", None, None, None, false, &config).unwrap();
+        assert_eq!(inside_filter(Some(true), net_list.as_slice(), &ip), true);
+        assert_eq!(inside_filter(Some(false), net_list.as_slice(), &ip), false);
+
+        let ip = parse_address_mask("192.168.1.0/24", None, None, None, false, &config).unwrap();
+        assert_eq!(inside_filter(Some(false), net_list.as_slice(), &ip), true);
+        assert_eq!(inside_filter(Some(true), net_list.as_slice(), &ip), false);
+
+        net_list
+            .push(parse_address_mask("192.168.1.0/24", None, None, None, false, &config).unwrap());
+
+        let ip = parse_address_mask("192.168.0.0/24", None, None, None, false, &config).unwrap();
+        assert_eq!(inside_filter(Some(true), net_list.as_slice(), &ip), true);
+        assert_eq!(inside_filter(Some(false), net_list.as_slice(), &ip), false);
+
+        let ip = parse_address_mask("192.168.1.0/24", None, None, None, false, &config).unwrap();
+        assert_eq!(inside_filter(Some(true), net_list.as_slice(), &ip), true);
+        assert_eq!(inside_filter(Some(false), net_list.as_slice(), &ip), false);
+    }
 }
